package de.ozgcloud.nachrichten.postfach.bayernid;

import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.stream.Stream;

import jakarta.annotation.PostConstruct;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.stereotype.Component;

import com.google.protobuf.ByteString;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.postfach.FileId;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachRemoteService;
import de.ozgcloud.nachrichten.postfach.PostfachRuntimeException;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.BayernIdProxyServiceGrpc.BayernIdProxyServiceStub;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcAttachments;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcSendBayernIdMessageRequest;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcSendBayernIdMessageResponse;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;
import lombok.extern.log4j.Log4j2;
import net.devh.boot.grpc.client.inject.GrpcClient;

@Log4j2
@ConditionalOnProperty(prefix = "ozgcloud.bayernid", name = { "enabled" }, havingValue = "true")
@Component
class BayernIdPostfachRemoteService implements PostfachRemoteService {

	static final String POSTFACH_TYPE = "BAYERN_ID";

	@GrpcClient("bayern-id")
	private BayernIdProxyServiceStub bayernIdProxyServiceStub;

	@Autowired
	private BayernIdProperties bayernIdProperties;
	@Autowired
	private BayernIdPostfachResponseHandler responseHandler;
	@Autowired
	private BayernIdPostfachNachrichtMapper mapper;
	@Autowired
	private BayernIdAttachmentService bayernIdAttachmentService;

	@PostConstruct
	void init() {
		LOG.info("BayernID Postfach remote service initialized.");
	}

	@Override
	public void sendMessage(PostfachNachricht nachricht) {
		try {
			var sender = createMessageWithFilesSender(nachricht).send();
			var response = waitUntilTransferCompleted(sender);
			responseHandler.handleResponse(mapper.fromSendBayernIdMessageResponse(response));
		} catch (BayernIdServerException e) {
			throw e;
		} catch (RuntimeException e) {
			throw new PostfachRuntimeException(e);
		}
	}

	MessageWithFilesSender createMessageWithFilesSender(PostfachNachricht nachricht) {
		return MessageWithFilesSender.builder()
				.reqObserverBuilder(buildCallStreamObserverBuilder())
				.messageMetadata(mapper.toSendBayernIdMessageMetadataRequest(nachricht, bayernIdProperties.getAbsender()))
				.attachmentIds(nachricht.getAttachments())
				.toAttachment(buildAttachmentBuilder())
				.attachmentMetadataMapper(buildAttachmentMetadataMapper())
				.chunkBuilder(buildChunkRequest())
				.build();
	}

	Function<StreamObserver<GrpcSendBayernIdMessageResponse>, CallStreamObserver<GrpcSendBayernIdMessageRequest>> buildCallStreamObserverBuilder() {
		return responseObserver ->
				(CallStreamObserver<GrpcSendBayernIdMessageRequest>) bayernIdProxyServiceStub.sendMessageAsStream(responseObserver);
	}

	Function<String, BayernIdAttachment> buildAttachmentBuilder() {
		return attachmentId -> bayernIdAttachmentService.getMessageAttachment(FileId.from(attachmentId));
	}

	BiFunction<byte[], Integer, GrpcSendBayernIdMessageRequest> buildChunkRequest() {
		return (bytes, length) -> GrpcSendBayernIdMessageRequest.newBuilder()
				.setAttachments(GrpcAttachments.newBuilder().setContent(ByteString.copyFrom(bytes, 0, length)).build())
				.build();
	}

	Function<BayernIdAttachment, GrpcSendBayernIdMessageRequest> buildAttachmentMetadataMapper() {
		return mapper::toSendBayernIdAttachmentsMetadataRequest;
	}

	GrpcSendBayernIdMessageResponse waitUntilTransferCompleted(MessageWithFilesSender sender) {
		try {
			return sender.getResultFuture().get(2, TimeUnit.MINUTES);
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
			sender.cancelOnError(e);
			throw new TechnicalException("Waiting for finishing upload was interrupted.", e);
		} catch (ExecutionException | TimeoutException e) {
			sender.cancelOnTimeout();
			throw new TechnicalException("Error / Timeout on uploading data.", e);
		}
	}

	@Override
	public Stream<PostfachNachricht> getAllMessages() {
		throw new UnsupportedOperationException("Bayern ID doesn't support this operation.");
	}

	@Override
	public void deleteMessage(String messageId) {
		throw new UnsupportedOperationException("Bayern ID doesn't support this operation.");
	}

	@Override
	public String getPostfachType() {
		return POSTFACH_TYPE;
	}

	@Override
	public boolean isReplyAllowed() {
		return false;
	}

}
